;
; Michael Rasmussen, 2017-03-07
; With improvements suggested here: http://eab.abime.net/showthread.php?t=86108
;

	include init.s

; must be kept one-frame
screenw=256
screenh=92*2

xstart=$81+(320-screenw)/2
ystart=$2c+(256-screenh)/2 ; first scanline
rowbytes=((screenw+15)/16)*2
bplbytes=(screenh/2)*rowbytes

blitpasssize=bplbytes/2

chunkyw=screenw/2
chunkyh=screenh/2

texshift=7
texturew=1<<texshift
textureh=1<<texshift

level3vec=(24+3)*4

copperlistwords=(screenh*4+40)*2 ; blah...

	SECTION mycode, CODE

; wait for line \1 (e.g. #$12d), note: must not be d0
; trashes d0, assumes a6=custom
waitline macro
.\@:
    	move.l  vposr(a6), d0
    	lsr.l	#8, d0
    	and.w	#$1ff, d0
    	cmp.w   \1, d0
    	bne.s   .\@
	endm

waitblit macro
	tst	dmaconr+custom		; Compatibility
.\@:
	btst.b	#DMAB_BLTDONE-8, dmaconr+custom
	bne.s	.\@
	endm

setcopptr macro
	move.w	\1, 6(\2)
	swap	\1
	move.w	\1, 2(\2)
	swap	\1
	add.l	#8, \2
        endm

        ; \1 = 0/1
setupcopper macro
        move.l  #copperlist\1, a0
        move.l  #copperlist, a1
        move.w  #copperlistwords-1, d0
        ; copy template
.\@:
        move.w  (a1)+, (a0)+
        dbf     d0, .\@

        ; setup blitter pointers for pass 0
        move.l  #copperlist\1, a0
        add.l   #bltpts0-copperlist, a0
	move.l	#chunkybuf\1, d0
        setcopptr d0, a0
	addq.l	#2, d0
        setcopptr d0, a0
	move.l	#screen+bplbytes, d0
        setcopptr d0, a0

        ; pass1
        move.l  #copperlist\1, a0
        add.l   #bltpts1-copperlist, a0
	move.l	#chunkybuf\1, d0
	add.l	#bplbytes*2-4, d0
        setcopptr d0, a0
	addq.l	#2, d0
        setcopptr d0, a0
	move.l	#screen+bplbytes-2, d0
        setcopptr d0, a0

	; setup bitplane pointers, d0 = bitplanes
	; setup bpl 0 -> plane buf 0
	;       bpl 1 -> plane buf 0
	;       bpl 2 -> plane buf 1
	;       bpl 3 -> plane buf 1
	;	bpl 4 -> plane buf 2
        move.l  #screen, d0
	move.l	#copperlist\1, a0
        add.l   #bplpts-copperlist, a0
        setcopptr d0, a0
        setcopptr d0, a0
	add.l	#bplbytes, d0
        setcopptr d0, a0
        setcopptr d0, a0
	add.l	#bplbytes, d0
        setcopptr d0, a0
        endm ; setupcopper

init:
        move.l  #-1, vblanks ; cheat and start the vblank counter at -1 ignoring the first frame

	move.l	#custom, a6

	; init texture pointer
        move.l	#bmp, a1
        move.w	10(a1), d0 ; grab lower bits of bfOffBits
        rol.w	#8, d0
        add.w	d0, a1
        move.l	a1, texture

	; invert bmp
	move.l	texture, a0
	move.l	texture, a1
	add.l	#texturew*(textureh-1), a1
	move.w	#textureh/2-1, d7
.texy:
	move.l	#texturew/4-1, d6
.texx:
	move.l	(a0), d0
	move.l	(a1), d1
	move.l	d1, (a0)+
	move.l	d0, (a1)+
	dbf	d6, .texx
	add.l	#-texturew*2, a1
	dbf	d7, .texy

	; scramble pixels in texture and repeat
	move.l 	#scrambledtex, a0
	move.l 	#scrambledtex+1*65536, a1
	move.l 	#scrambledtex+2*65536, a2
	move.l 	#scrambledtex+3*65536, a3
	move.w	#32768/(texturew*textureh)-1, d6
.repeat:
	move.w	#texturew*textureh-1, d7
	move.l 	texture, a4
.scramble:
	moveq	#0, d0
	move.b	(a4)+, d0
	lsl.w	#6, d0
	lsr.b	#6, d0
	lsl.w	#6, d0
	move.w	d0, (a0)+
	lsr.w	#2, d0
	move.w	d0, (a1)+
	lsr.w	#2, d0
	move.w	d0, (a2)+
	lsr.w	#2, d0
	move.w	d0, (a3)+
	dbf	d7, .scramble
	dbf	d6, .repeat

	move.l	#scrambledtex, texture ; point texture at scrambled buffer

	; iniitalize mask in bpl 3 (== 5)
	move.l	#screen+2*bplbytes, a0
	move.l	#$aaaaaaaa, d0
	move.w	#bplbytes/4-1, d1
.fill:
	move.l	d0, (a0)+
	dbf	d1, .fill

	;
	; Wait for vblank
	;
	waitline #ystart+screenh

	;
	; Setup palette from image
	;
	move.l	#bmp+$36, a1 ; color table (40 bytes + $e initial)
	move.l	#custom+color, a0 ; start from color 0
	moveq	#16-1, d7	; 16 color palette
.col:
	; (a1).l = Bb Gg Rr ??
	move.l	(a1)+, d0	; d0 = Bb Gg Rr ??
	lsr.w	#4, d0		; d0 = Bb Gg 0R r?
	move.l	d0, d1		; d1 = Bb Gg 0R r?
	swap	d0		; d0 = 0R r? Bb Gg
	and.w	#$f0f0, d0	; d0 = 0R r? B0 G0
	move.w	d0, d2		; d2 = ?? ?? B0 G0
	rol.w	#4, d2		; d2 = ?? ?? 0G 0B
	or.b	d2, d0		; d0 = 0R r? B0 GB
	move.b	d0, d1		; d1 = Bb Gg 0R GB
	cmp.w	#16-1, d7
	bne.s	.notfirst
	move.w	d1, d6 ; save background color
.notfirst:
	move.w	d1, (a0)+
	dbf	d7, .col

	; set color 16-31 (mask) to background color
	move.w	#16-1, d7
.setrest:
	move.w	d6, (a0)+
	dbf	d7, .setrest

        ; init copper lists
        setupcopper 0
        setupcopper 1

	move.w	#(5<<12)!$0200, d0  ; $200=COLOR
	tst.b	agapresent
	bne.s	.setcon0
	add.w	#(2<<12), d0	    	; set "7-bplmode", OCS/ECS only
	move.w	#$aaaa, bpldat+4*2(a6) 	; make sure data registers are
	move.w	#$0000, bpldat+5*2(a6) 	; loaded with the correct masks
.setcon0:
	move.w	d0, bplcon0(a6)
	move.w	#$0000, bplcon1(a6)
	move.w	#$0000, bpl1mod(a6)
	move.w	#$0000, bpl2mod(a6)
	move.w 	#(ystart<<8)!xstart, diwstrt(a6)
	move.w 	#(((ystart+screenh)<<8)&$ff00)!((xstart+screenw)&$ff), diwstop(a6)
	move.w	#xstart/2-8, ddfstrt(a6)
	move.w	#xstart/2-8+8*((screenw+15)/16-1), ddfstop(a6)

	; common blitter initialization
	moveq	#-1, d0
	move.l	d0, bltafwm(a6)
	move.w	#$ff00, bltcdat(a6)
	move.w	#2, d0
	move.w	d0, bltamod(a6)
	move.w	d0, bltbmod(a6)
	move.w	#0, bltdmod(a6)

        move.w  #2, copcon(a6) ; allow the copper to do dangerous stuff
	; Start copper
	move.l	#copperlist0, cop1lc(a6)

	; Setup DMA
	move.w 	#DMAF_SETCLR!DMAF_MASTER!DMAF_RASTER!DMAF_COPPER!DMAF_BLITTER, dmacon(a6)

	; save level3 vector
	move.l	vectorbase, a0
	move.l	level3vec(a0), oldlev3int
	; install new level3 vector
	move.l	#lev3int, level3vec(a0)
	; enable interrupts
	move.w	#INTF_SETCLR!INTF_INTEN!INTF_VERTB, intena(a6)
	rts

cleanup:
	move.l	#custom, a6
	; disable interrupts
	move.w	#$7fff, intena(a6) ; disable interrupts
	; restore level3 vector
	move.l	vectorbase, a0
	move.l	oldlev3int, level3vec(a0)
	rts

dosname: dc.b 'dos.library',0
msgfmt:
        dc.b 'framenum=$%lx vblanks=$%lx vblanks/framenum=$%x.%04x', 13, 10, 'Thanks for watching!', 13, 10, 0
msgbuf: ds.b 256
        even
msglen:  ds.l 1
fmtargs:
        ds.l 3

dosbase: ds.l 1
stdout:  ds.l 1

msg_putch:
        addq.l  #1, msglen
        move.b  d0, (a3)+
        rts

atexit:
        move.l  vblanks, d1
        swap    d1
        clr.w   d1
        moveq   #0, d7
        move.l  framenum, d2
        beq.s   .skipdiv ; division by zero
        ; stupid division...
        bra.s   .check
.div:
        addq.l  #1, d7
        sub.l   d2, d1
.check:
        cmp.l   d2, d1
        bge.s   .div
        ; d7 = (vblanks<<16)/framenum
.skipdiv:

; open dos.library
        move.l	execbase, a6
	lea	dosname, a1		; a1 = name
	moveq	#0, d0			; d0 = version (0 = any)
	jsr	_LVOOpenLibrary(a6)
	move.l	d0, a6
	move.l	d0, dosbase
        ; get stdout
        jsr     _LVOOutput(a6)
        move.l  d0, stdout

        move.l  framenum, fmtargs
        move.l  vblanks, fmtargs+4
        move.l  d7, fmtargs+8

        move.l  #msgfmt, a0
        move.l  #fmtargs, a1
        move.l  #msg_putch, a2
        move.l  #msgbuf, a3
        move.l  execbase, a6
        jsr     _LVORawDoFmt(a6)

        move.l  stdout, d1
        move.l  #msgbuf, d2
        move.l  msglen, d3
        move.l  dosbase, a6
        jsr     _LVOWrite(a6)

        ; close dos.library
        move.l	dosbase, a1
	move.l	execbase, a6
	jsr	_LVOCloseLibrary(a6)

        moveq   #0, d0 ; d0 = exit code
        rts

lev3int:
	; check cause
	btst.b	#INTB_VERTB, intreqr+custom+1
	beq.b	.novertb
        addq.l  #1, vblanks
	move.w	#INTF_VERTB, intreq+custom ; acknowledge interrupt
	move.w	#INTF_VERTB, intreq+custom ; twice for compatability
.novertb:
	rte

mainloop:
	move.l	#custom, a6
	btst.b	#2, potinp(a6)
	bne.s	.normb
	rts
.normb:

        ; Do effect
	move.l	#chunkybuf0, a0
        btst    #0, framenum+3
        bne.s   .doeffect
        move.l  #chunkybuf1, a0
.doeffect:
	move.l	texture, a1
	move.l	framenum, d4
	move.w	d4, d0
	lsl.w	#8, d4
	add.w	d0, d4

	and.w	#(textureh-1)*texturew+(texturew-1), d4
	add.w	d4, d4 ; needs to be double since the tundat is...
	add.w	d4, a1 ; offset texture
        moveq   #1, d0
        swap    d0     ; d0 = $10000
	move.l	a1, a2
	add.l	d0, a2
	move.l	a2, a3
	add.l	d0, a3
	move.l	a3, a4
	add.l	d0, a4

	jsr	unrolled_code

        ; Switch to next copper list
        move.l  #copperlist0, d7
        btst    #0, framenum+3
        bne.s   .setcop
        move.l  #copperlist1, d7
.setcop:

        ; make sure the blitter is done with the last chunky buffer
.waitready:
        btst    #INTB_COPER, intreqr+1(a6)
        beq.s   .waitready

	move.l	d7, cop1lc(a6)

        add.l   #1, framenum

	rts


	SECTION mychipbss, BSS_C
screen:         ds.w bplbytes*3/2
chunkybuf0:     ds.w bplbytes*2/2
chunkybuf1:     ds.w bplbytes*2/2
copperlist0:    ds.w copperlistwords
copperlist1:    ds.w copperlistwords

	SECTION mybss, BSS
framenum:       ds.l 1
vblanks:        ds.l 1
texture:        ds.l 1
oldlev3int:     ds.l 1
	cnop 0, 4
scrambledtex:   ds.w 32768*4

	SECTION mydata, DATA
bmp:
        incbin "mess128x128x4.bmp"

	cnop 0, 4
unrolled_code:
        if (chunkyw=128)&(chunkyh=88)
        incbin unrolled_code_128x88.bin
        endc
        if (chunkyw=128)&(chunkyh=92)
        incbin unrolled_code_128x92.bin
        endc
        if (chunkyw=128)&(chunkyh=96)
        incbin unrolled_code_128x96.bin
        endc
        if (chunkyw=160)&(chunkyh=128)
        incbin unrolled_code_160x128.bin
        endc

copperlist:
        ; setup first blitter pass
	dc.w	bltcon0,        SRCA!SRCB!DEST!$E4
	dc.w	bltcon1,        (8<<12)
bltpts0:
        dc.w    bltapt+$00,     $0000
        dc.w    bltapt+$02,     $0000
        dc.w    bltbpt+$00,     $0000
        dc.w    bltbpt+$02,     $0000
        dc.w    bltdpt+$00,     $0000
        dc.w    bltdpt+$02,     $0000
        ; start blitter
        dc.w    bltsize, 1 ;  (1024x1 word blit)
bplpts:
	; bitplane pointer setup
	dc.w bplpt+$00, $0000 ; 0
	dc.w bplpt+$02, $0000
	dc.w bplpt+$04, $0000 ; 1
	dc.w bplpt+$06, $0000
	dc.w bplpt+$08, $0000 ; 2
	dc.w bplpt+$0A, $0000
	dc.w bplpt+$0C, $0000 ; 3
	dc.w bplpt+$0E, $0000
	dc.w bplpt+$10, $0000 ; 4
	dc.w bplpt+$12, $0000 
	dc.w bplpt+$14, $0000 ; 5
	dc.w bplpt+$16, $0000

        ; wait for blit and continue pass 0
	dc.w $0001, $0000
        dc.w bltsize, ((blitpasssize&1023)<<6)!1
        ; wait for blit and do pass 1
	dc.w $0001, $0000
	dc.w bltcon0,        (8<<12)!SRCA!SRCB!DEST!$E4
	dc.w bltcon1,        BLITREVERSE
bltpts1:
        dc.w bltapt+$00,     $0000
        dc.w bltapt+$02,     $0000
        dc.w bltbpt+$00,     $0000
        dc.w bltbpt+$02,     $0000
        dc.w bltdpt+$00,     $0000
        dc.w bltdpt+$02,     $0000

        ; start blitter
        dc.w bltsize, 1 ;  (1024x1 word blit)
        ; wait for blit and continue pass 1
	dc.w $0001, $0000
        dc.w bltsize, ((blitpasssize&1023)<<6)!1

        ; wait for blitter to finish and set copper interrupt request bit
	dc.w $0001, $0000
        dc.w intreq, INTF_SETCLR!INTF_COPER
        ; hope blitter is finished in time ;)
quaterscreen=bplbytes/4
.y set ystart-1
	rept screenh/2
	dc.w ((.y&$ff)<<8)!$df, -2
        ; repeat
	dc.w bplcon1, (2<<4)!1
	dc.w bpl1mod, -rowbytes
	dc.w bpl2mod, -rowbytes
	dc.w (((.y+1)&$ff)<<8)!$df, -2
	dc.w bplcon1, 1<<4
        ; move to next scanline
        if (((.y+1)-ystart)/2)&3=3
	dc.w bpl1mod, -quaterscreen*3
	dc.w bpl2mod, -quaterscreen*3
        else
	dc.w bpl1mod, quaterscreen-rowbytes
	dc.w bpl2mod, quaterscreen-rowbytes
        endc
.y set .y+2
	endr ; screenh

        ; could wait for a later screen position here..
        ; clear copper interrupt reuest bit (= no more copper updates allowed)
        dc.w intreq, INTF_COPER
	; end of copper list
	dc.l -2

        if *-copperlist>copperlistwords*2
        error
        endc
